/** @page tut_bson Working with BSON objects
 *
 * In this section, we'll cover the basics of working with BSON
 * objects in a few big steps. Working with BSON is fairly
 * straightforward, so we will not be going into much details here.
 *
 * Contents:
 *  - @subpage tut_bson_build
 *  - @subpage tut_bson_traverse
 */
