/** @page tut_mongo_sync Working with the Mongo Sync API
 *
 * In this section we'll be going over various parts of the
 * synchronous API provided by libmongo-client. From connecting to a
 * single host, through replica sets to performing various more
 * complex operations.
 *
 * Contents:
 *  - @subpage tut_mongo_sync_connect
 *  - @subpage tut_mongo_sync_insert
 *  - @subpage tut_mongo_sync_query
 *  - @subpage tut_mongo_sync_query_complex
 *  - @subpage tut_mongo_sync_cmd_create
 *  - @subpage tut_mongo_sync_cmd_index_create
 *  - @subpage tut_mongo_sync_cmd_custom
 */
